import distutils
import distutils.version
from functools import reduce
from operator import xor

import semantic_version
import six
from docutils import nodes
from semantic_version import Spec


semver_version = distutils.version.StrictVersion(semantic_version.__version__)

if semver_version < distutils.version.StrictVersion('2.7.0'):

    StrictVersion = semantic_version.StrictVersion

    class Version(StrictVersion):
        """
        Version subclass toggling ``partial=True`` by default.
        """
        def __init__(self, version_string, partial=True):
            super(Version, self).__init__(version_string, partial)

else:

    class Version(semantic_version.Version):
        """
        Version subclass. Ignore partial.
        """
        def __init__(self, version_string, partial=True):
            super(Version, self).__init__(version_string)



# Issue type list (keys) + color values
ISSUE_TYPES = {
    'bug': 'A04040',
    'feature': '40A056',
    'support': '4070A0',
}


class Issue(nodes.Element):
    # Technically, we just need number, but heck, you never know...
    _cmp_keys = ('type', 'number', 'backported', 'major')

    @property
    def type(self):
        return self['type_']

    @property
    def is_featurelike(self):
        if self.type == 'bug':
            return self.major
        else:
            return not self.backported

    @property
    def is_buglike(self):
        return not self.is_featurelike

    @property
    def backported(self):
        return self.get('backported', False)

    @property
    def major(self):
        return self.get('major', False)

    @property
    def number(self):
        return self.get('number', None)

    @property
    def spec(self):
        return self.get('spec', None)

    def __eq__(self, other):
        for attr in self._cmp_keys:
            if getattr(self, attr, None) != getattr(other, attr, None):
                return False
        return True

    def __hash__(self):
        return reduce(xor, [hash(getattr(self, x)) for x in self._cmp_keys])

    def minor_releases(self, manager):
        """
        Return all minor release line labels found in ``manager``.
        """
        # TODO: yea deffo need a real object for 'manager', heh. E.g. we do a
        # very similar test for "do you have any actual releases yet?"
        # elsewhere. (This may be fodder for changing how we roll up
        # pre-major-release features though...?)
        return [
            key for key, value in six.iteritems(manager)
            if any(x for x in value if not x.startswith('unreleased'))
        ]

    def default_spec(self, manager):
        """
        Given the current release-lines structure, return a default Spec.

        Specifics:

        * For feature-like issues, only the highest major release is used, so
          given a ``manager`` with top level keys of ``[1, 2]``, this would
          return ``Spec(">=2")``.

            * When ``releases_always_forwardport_features`` is ``True``, that
              behavior is nullified, and this function always returns the empty
              ``Spec`` (which matches any and all versions/lines).

        * For bugfix-like issues, we only consider major release families which
          have actual releases already.

            * Thus the core difference here is that features are 'consumed' by
              upcoming major releases, and bugfixes are not.

        * When the ``unstable_prehistory`` setting is ``True``, the default
          spec starts at the oldest non-zero release line. (Otherwise, issues
          posted after prehistory ends would try being added to the 0.x part of
          the tree, which makes no sense in unstable-prehistory mode.)
        """
        # TODO: I feel like this + the surrounding bits in add_to_manager()
        # could be consolidated & simplified...
        specstr = ""
        # Make sure truly-default spec skips 0.x if prehistory was unstable.
        stable_families = manager.stable_families
        if manager.config.releases_unstable_prehistory and stable_families:
            specstr = ">={}".format(min(stable_families))
        if self.is_featurelike:
            # TODO: if app->config-><releases_always_forwardport_features or
            # w/e
            if True:
                specstr = ">={}".format(max(manager.keys()))
        else:
            # Can only meaningfully limit to minor release buckets if they
            # actually exist yet.
            buckets = self.minor_releases(manager)
            if buckets:
                specstr = ">={}".format(max(buckets))
        return Spec(specstr) if specstr else Spec()

    def add_to_manager(self, manager):
        """
        Given a 'manager' structure, add self to one or more of its 'buckets'.
        """
        # Derive version spec allowing us to filter against major/minor buckets
        spec = self.spec or self.default_spec(manager)
        # Only look in appropriate major version/family; if self is an issue
        # declared as living in e.g. >=2, this means we don't even bother
        # looking in the 1.x family.
        families = [Version(str(x)) for x in manager]
        versions = list(spec.filter(families))
        for version in versions:
            family = version.major
            # Within each family, we further limit which bugfix lines match up
            # to what self cares about (ignoring 'unreleased' until later)
            candidates = [
                Version(x)
                for x in manager[family]
                if not x.startswith('unreleased')
            ]
            # Select matching release lines (& stringify)
            buckets = []
            bugfix_buckets = [str(x) for x in spec.filter(candidates)]
            # Add back in unreleased_* as appropriate
            # TODO: probably leverage Issue subclasses for this eventually?
            if self.is_buglike:
                buckets.extend(bugfix_buckets)
                # Don't put into JUST unreleased_bugfix; it implies that this
                # major release/family hasn't actually seen any releases yet
                # and only exists for features to go into.
                if bugfix_buckets:
                    buckets.append('unreleased_bugfix')
            # Obtain list of minor releases to check for "haven't had ANY
            # releases yet" corner case, in which case ALL issues get thrown in
            # unreleased_feature for the first release to consume.
            # NOTE: assumes first release is a minor or major one,
            # but...really? why would your first release be a bugfix one??
            no_releases = not self.minor_releases(manager)
            if self.is_featurelike or self.backported or no_releases:
                buckets.append('unreleased_feature')
            # Now that we know which buckets are appropriate, add ourself to
            # all of them. TODO: or just...do it above...instead...
            for bucket in buckets:
                manager[family][bucket].append(self)

    def __repr__(self):
        flag = ''
        if self.backported:
            flag = 'backported'
        elif self.major:
            flag = 'major'
        elif self.spec:
            flag = self.spec
        if flag:
            flag = ' ({})'.format(flag)
        return '<{issue.type} #{issue.number}{flag}>'.format(issue=self,
                                                             flag=flag)


class Release(nodes.Element):
    @property
    def number(self):
        return self['number']

    @property
    def minor(self):
        # TODO: use Version
        return '.'.join(self.number.split('.')[:-1])

    @property
    def family(self):
        # TODO: use Version.major
        # TODO: and probs just rename to .major, 'family' is dumb tbh
        return int(self.number.split('.')[0])

    def __repr__(self):
        return '<release {}>'.format(self.number)



import sys

from .metadata import get_package_version
from .logger import get_logger


__version__ = get_package_version(path=__file__, package_name='sdsstools')

NAME = 'sdsstools'


# Get a logger, mostly for the nice colouring.
log = get_logger(NAME)


from .configuration import *
from .logger import *
from .metadata import *
from ._vendor import releases, toml

# Allow to access releases as sdsstools.releases. This is important to
# define it as an extension in Sphinx.
sys.modules['sdsstools.releases'] = releases



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-08
# @Filename: configuration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import inspect
import itertools
import os
import pathlib

import yaml


__all__ = ['read_yaml_file', 'merge_config', 'get_config']


def read_yaml_file(path):
    """Read a YAML file and returns a dictionary."""

    if isinstance(path, (str, pathlib.Path)):
        with open(path, 'r') as fp:
            config = yaml.safe_load(fp)
    else:
        # Assume it's an stream
        config = yaml.safe_load(path)

    return config


def merge_config(user, default):
    """Merges a user configuration with the default one."""

    if isinstance(user, dict) and isinstance(default, dict):
        for kk, vv in default.items():
            if kk not in user:
                user[kk] = vv
            else:
                user[kk] = merge_config(user[kk], vv)

    return user


def get_config(name, config_file=None, allow_user=True, user_path=None,
               config_envvar=None, merge_mode='update'):
    """Returns a configuration dictionary.

    The configuration dictionary is created by merging the default
    configuration file that is part of the library (normally in
    ``etc/<name>.yml``) with a user configuration file. The path to the user
    configuration file can be defined as an environment variable to be passed
    to this function in ``config_envvar`` or as a path in ``user_path``.
    The environment variable, if exists, always takes precedence.

    Parameters
    ----------
    name : str
        The name of the package.
    config_file : str
        The path to the configuration file. If `None`, defaults to
        ``etc/<name>.yml`` relative to the file that called `.get_config`.
    allow_user : bool
        If `True`, looks for an user configuration file and merges is to the
        default configuration. Otherwise it returns just the default
        configuration.
    user_path : str
        The path to the user configuration file. Defaults to
        ``~/.config/sdss/<name>/<name>.yml``. Ignored if the file does not
        exist.
    config_envvar : str
        The environment variable that contains the path to the user
        configuration file. Defaults to ``<name>_CONFIG_PATH``. If the
        environment variable exists, the ``user_path`` is ignored.
    merge_mode : str
        Defines how the default and user dictionaries will be merged. If
        ``update``, the user dictionary will be used to update the default
        configuration. If ``replace``, only the user configuration will be
        returned.

    Returns
    -------
    config : dict
        A dictionary containing the configuration.

    """

    assert merge_mode in ['update', 'replace'], 'invalid merge mode.'

    if not config_file:
        try:
            frame = inspect.stack()[1]
            module = inspect.getmodule(frame[0])
            dirname = os.path.dirname(module.__file__)
            config_file = os.path.join(dirname, f'etc/{name}.yml')
        except AttributeError:
            config_file = None

    if config_file and os.path.exists(config_file):
        config = read_yaml_file(config_file)
    else:
        config = {}

    if allow_user is False:
        return config

    config_envvar = config_envvar or '{}_CONFIG_PATH'.format(name.upper())

    if user_path is not None:
        user_path = os.path.expanduser(os.path.expandvars(user_path))
    else:
        # Test a few default paths and exit when finds one.
        default_paths = [f'~/.config/sdss/{name}',
                         f'~/.config/sdss/{name}/{name}',
                         f'~/.{name}/{name}']
        extensions = ['.yaml', '.yml']
        for path, extension in itertools.product(default_paths, extensions):
            user_path = os.path.expanduser(path + extension)
            if os.path.exists(user_path):
                break

    if config_envvar in os.environ:
        custom_config_fn = os.environ[config_envvar]
    elif os.path.exists(user_path):
        custom_config_fn = user_path
    else:
        return config

    user_config = read_yaml_file(custom_config_fn) or {}

    if merge_mode == 'update':
        return merge_config(user_config, config)
    else:
        return user_config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2017-10-11
# @Filename: logger.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import copy
import logging
import os
import re
import shutil
import sys
import traceback
import warnings
from logging.handlers import TimedRotatingFileHandler

from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import get_lexer_by_name

from .color_print import color_text


__all__ = ['get_logger']


def get_exception_formatted(tp, value, tb):
    """Adds colours to tracebacks."""

    tbtext = ''.join(traceback.format_exception(tp, value, tb))
    lexer = get_lexer_by_name('pytb', stripall=True)
    formatter = TerminalFormatter()
    return highlight(tbtext, lexer, formatter)


def colored_formatter(record):
    """Prints log messages with colours."""

    colours = {'info': 'blue',
               'debug': 'magenta',
               'warning': 'yellow',
               'critical': 'red',
               'error': 'red'}

    levelname = record.levelname.lower()

    if levelname.lower() in colours:
        levelname_color = colours[levelname]
        header = color_text('[{}]: '.format(levelname.upper()),
                            levelname_color)
    else:
        header = f'[{levelname}]'

    message = record.getMessage()

    if levelname == 'warning':
        warning_category_groups = re.match(r'^.*?\s*?(\w*?Warning): (.*)', message)
        if warning_category_groups is not None:
            warning_category, warning_text = warning_category_groups.groups()

            # Temporary ignore warnings from pymodbus. The normal warnings.simplefilter
            # does not work because pymodbus forces them to show.
            if re.match('"@coroutine" decorator is deprecated.+', warning_text):
                return

            warning_category_colour = color_text('({})'.format(warning_category), 'cyan')
            message = '{} {}'.format(color_text(warning_text, ''), warning_category_colour)

    sys.__stdout__.write('{}{}\n'.format(header, message))
    sys.__stdout__.flush()

    return


class SDSSFormatter(logging.Formatter):
    """Custom `Formatter <logging.Formatter>`."""

    base_fmt = '%(asctime)s - %(levelname)s - %(message)s'
    ansi_escape = re.compile(r'\x1b[^m]*m')

    def __init__(self, fmt=base_fmt):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):

        # Copy the record so that any modifications we make do not
        # affect how the record is displayed in other handlers.
        record_cp = copy.copy(record)

        record_cp.msg = self.ansi_escape.sub('', record_cp.msg)

        # The format of a warnings redirected with warnings.captureWarnings
        # has the format <path>: <category>: message\n  <some-other-stuff>.
        # We reorganise that into a cleaner message. For some reason in this
        # case the message is in record.args instead of in record.msg.
        if record_cp.levelno == logging.WARNING and len(record_cp.args) > 0:
            match = re.match(r'^(.*?):\s*?(\w*?Warning): (.*)', record_cp.args[0])
            if match:
                message = '{1} - {2} [{0}]'.format(*match.groups())
                record_cp.args = tuple([message] + list(record_cp.args[1:]))

        return logging.Formatter.format(self, record_cp)


class SDSSLogger(logging.Logger):
    """Custom logging system.

    Parameters
    ----------
    name : str
        The name of the logger.
    log_level : int
        The initial logging level for the console handler.
    capture_warnings : bool
        Whether to capture warnings and redirect them to the log.

    """

    def __init__(self, name):

        # Placeholder for the last error-level message emitted.
        self._last_error = None

        super(SDSSLogger, self).__init__(name)

    def init(self, log_level=logging.INFO, capture_warnings=True):

        # Set levels
        self.setLevel(logging.DEBUG)

        # Sets the console handler
        self.sh = logging.StreamHandler()
        self.sh.emit = colored_formatter
        self.addHandler(self.sh)
        self.sh.setLevel(log_level)

        # Placeholders for the file handler.
        self.fh = None
        self.log_filename = None

        # A header that precedes every message.
        self.header = ''

        # Catches exceptions
        sys.excepthook = self._catch_exceptions

        self.warnings_logger = None

        if capture_warnings:
            self.capture_warnings()

    def _catch_exceptions(self, exctype, value, tb):
        """Catches all exceptions and logs them."""

        self.error(get_exception_formatted(exctype, value, tb))

    def capture_warnings(self):
        """Capture warnings.

        When `logging.captureWarnings` is `True`, all the warnings are
        redirected to a logger called ``py.warnings``. We add our handlers
        to the warning logger.

        """

        logging.captureWarnings(True)

        self.warnings_logger = logging.getLogger('py.warnings')

        # Only enable the sh handler if none is attached to the warnings
        # logger yet. Prevents duplicated prints of the warnings.
        for handler in self.warnings_logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                return

        self.warnings_logger.addHandler(self.sh)

    def asyncio_exception_handler(self, loop, context):
        """Handle an uncaught asyncio exception and reports it."""

        exception = context.get('exception', None)
        if exception:
            try:
                raise exception
            except Exception:
                self.exception(context['message'])
        else:
            message = context['message'].splitlines()
            for line in message:
                self.error(f'{line}')

    def save_log(self, path):
        shutil.copyfile(self.log_filename, os.path.expanduser(path))

    def start_file_logger(self, path, log_level=logging.DEBUG, mode='a', rotating=True):
        """Start file logging."""

        log_file_path = os.path.expanduser(path)
        logdir = os.path.dirname(log_file_path)

        try:

            if not os.path.exists(logdir):
                os.makedirs(logdir)

            if rotating:
                self.fh = TimedRotatingFileHandler(str(log_file_path),
                                                   when='midnight',
                                                   utc=True)
                self.fh.suffix = '%Y-%m-%d_%H:%M:%S'
            else:
                self.fh = logging.FileHandler(str(log_file_path), mode=mode)

        except (IOError, OSError) as ee:

            warnings.warn('log file {0!r} could not be opened for '
                          'writing: {1}'.format(log_file_path, ee),
                          RuntimeWarning)

        else:

            self.fh.setFormatter(SDSSFormatter())
            self.addHandler(self.fh)
            self.fh.setLevel(log_level)

            if self.warnings_logger:
                self.warnings_logger.addHandler(self.fh)

            self.log_filename = log_file_path

    def handle(self, record):
        """Handles a record but first stores it."""

        if hasattr(self, 'header') and self.header is not None:
            record.msg = self.header + record.msg

        if record.levelno == logging.ERROR:
            self._last_error = record.getMessage()

        return super().handle(record)

    def get_last_error(self):
        """Returns the last error emitted."""

        return self._last_error

    def set_level(self, level):
        """Sets levels for both sh and (if initialised) fh."""

        self.sh.setLevel(level)

        if self.fh:
            self.fh.setLevel(level)


def get_logger(name, **kwargs):
    """Gets a new logger."""

    orig_logger = logging.getLoggerClass()

    logging.setLoggerClass(SDSSLogger)

    log = logging.getLogger(name)
    log.init(**kwargs)

    logging.setLoggerClass(orig_logger)

    return log



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-12-30
# @Filename: metadata.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import configparser
import pathlib

import pkg_resources

from ._vendor import toml


METADATA_FILES = ['pyproject.toml', 'setup.cfg', 'setup.py']


def get_metadata_files(path):
    """Finds the list of metadata files for a package.

    Returns the path of the ``pyproject.toml``, ``setup.cfg``, or ``setup.py``
    that defines the metadata for the project.

    Parameters
    ----------
    path : str
        The path relative to which to search for metadata files.

    Returns
    -------
    path
        The path to the metadata file, or `None` if none was found.

    """

    path = pathlib.Path(path).absolute()
    if path.is_file():
        path = path.parent

    for mf in METADATA_FILES:
        if (path / mf).exists():
            return str(path / mf)

    # Walk up directories
    for parent in path.parents:
        for mf in METADATA_FILES:
            if (parent / mf).exists():
                return str(parent / mf)

    return None


def get_package_version(path=None, package_name=None, pep_440=False):
    """Returns the version of a package.

    First tries to determine if a metadata file is available and parses it.
    Otherwise, tries to use the egg/wheel information if the package has
    been installed.

    Parameters
    ----------
    path : str
        The path relative to which to search for metadata files.
    package_name : str
        The name of the package.
    pep_440 : bool
        If `True`, normalises the version string according to PEP 440.

    Returns
    -------
    version : str
        The version string, or `None` if it cannot be found.

    """

    assert path or package_name, 'either path or package_name are needed.'

    version = None

    if path:
        metadata_file = get_metadata_files(path)
        if metadata_file:
            if 'pyproject.toml' in str(metadata_file):
                pyproject = toml.load(open(metadata_file))
                if 'tool' in pyproject and 'poetry' in pyproject['tool']:
                    version = pyproject['tool']['poetry']['version']
            elif 'setup.cfg' in str(metadata_file):
                setupcfg = configparser.ConfigParser()
                setupcfg.read(metadata_file)
                if (setupcfg.has_section('metadata') and
                        setupcfg.has_option('metadata', 'version')):
                    version = setupcfg.get('metadata', 'version')

    if package_name:
        try:
            version = pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            return None

    if version and pep_440:
        version = str(pkg_resources.packaging.version.Version(version))

    return version



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-12-30
# @Filename: cli.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import os

import yaml

from sdsstools import __version__
from sdsstools.metadata import get_metadata_files


def main():
    """Invokes the tasks defined in _tasks.py, with some configuration."""

    try:
        from invoke import Collection, Program
        from sdsstools import _tasks
    except (ImportError, ModuleNotFoundError):
        raise ImportError('Cannot find invoke. Make sure sdsstools '
                          'is installed for development.')

    # Use the metadata file to determine the root of the package.
    metadata_file = get_metadata_files('.')
    if metadata_file is None:
        raise RuntimeError('cannot find the root of the package.')

    os.chdir(os.path.dirname(metadata_file))

    # Override the configuration if there is an invoke.yaml file next to the
    # metadata file.
    if os.path.exists('./invoke.yaml'):
        config = yaml.safe_load(open('invoke.yaml'))
        print('Using configuration file invoke.yaml')
    else:
        config = None

    program = Program(version=__version__,
                      namespace=Collection.from_module(_tasks, config=config))
    program.run()


if __name__ == '__main__':

    main()



# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date: 2017-09-27
# @Filename: tasks.py
# @License: BSD 3-Clause
# @Copyright: Brian Cherinka

import os

from invoke import Collection, task


# This file contains tasks that can be easily run from the shell terminal using the Invoke
# python package. If you do not have invoke, install it with pip install
# To list the tasks available, type invoke --list from the top-level repo directory

@task
def clean_docs(ctx, target=None):
    """Cleans up the docs."""

    if target is None:
        target = ctx.sphinx.target

    print('Cleaning the docs')
    ctx.run(f'rm -rf {target}/_build')


@task
def build_docs(ctx, target=None, clean=False):
    """Builds the Sphinx docs"""

    if target is None:
        target = ctx.sphinx.target

    if clean:
        print('Cleaning the docs')
        ctx.run(f'rm -rf {target}/_build', pty=True)

    print('Building the docs')
    os.chdir(f'{target}')
    ctx.run('make html', pty=True)


@task
def show_docs(ctx, target=None):
    """Shows the Sphinx docs"""

    if target is None:
        target = ctx.sphinx.target

    print('Showing the docs')
    os.chdir(f'{target}/_build/html')
    ctx.run('open ./index.html')


@task
def clean(ctx):
    """Cleans up build files and test files."""

    print('Cleaning')
    ctx.run('rm -rf htmlcov **/htmlcov .coverage* **/.coverage*')
    ctx.run('rm -rf build')
    ctx.run('rm -rf dist')
    ctx.run('rm -rf **/*.egg-info *.egg-info')


@task(clean)
def deploy(ctx, test=False):
    """Deploy the project to PyPI"""

    if test is False:
        print('Deploying to PyPI!')
        repository_url = ''
    else:
        print('Deploying to Test PyPI!')
        repository_url = '--repository-url https://test.pypi.org/legacy/'

    ctx.run('python setup.py sdist bdist_wheel')
    ctx.run(f'twine upload {repository_url} dist/*')


@task(name='install-deps')
def install_deps(ctx, extras=None):
    """Install only dependencies from setup.cfg."""

    import setuptools

    if not os.path.exists('setup.cfg'):
        raise RuntimeError('setup.cfg cannot be found. If your project uses '
                           'requirement files use pip install -r instead.')

    if extras:
        extras = extras.split(',')
    else:
        extras = []

    config = setuptools.config.read_configuration('setup.cfg')

    if not config['options']:
        return

    options = config['options']

    setup_requires = options.get('setup_requires', [])
    install_requires = options.get('install_requires', [])

    requires = setup_requires + install_requires
    requires_str = (' '.join('"' + item + '"' for item in requires))
    if len(requires) > 0:
        ctx.run(f'pip install --upgrade {requires_str}', pty=True)

    for extra in extras:
        print(f'Installing extras={extra}')
        if 'extras_require' not in options:
            raise RuntimeError('extras_require is not defined')
        extra_deps = options['extras_require'].get(extra, [])
        if len(extra_deps) > 0:
            extra_deps_str = (' '.join('"' + item + '"' for item in extra_deps))
            ctx.run(f'pip install --upgrade {extra_deps_str}', pty=True)


# create a collection of tasks
ns = Collection(clean, deploy, install_deps)

# create a sub-collection for the doc tasks
docs = Collection('docs')
docs.add_task(build_docs, 'build')
docs.add_task(clean_docs, 'clean')
docs.add_task(show_docs, 'show')
ns.add_collection(docs)

ns.configure({'sphinx': {'target': 'docs/sphinx'}})



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sdsstools',
 'sdsstools._vendor',
 'sdsstools._vendor.releases',
 'sdsstools._vendor.toml']

package_data = \
{'': ['*']}

install_requires = \
['docutils>=0.16,<0.17',
 'pygments>=2.5.2,<3.0.0',
 'pyyaml>=4.0',
 'semantic-version>=2.8.4,<3.0.0',
 'six>=1.14.0,<2.0.0']

extras_require = \
{'dev': ['invoke>=1.3.0,<2.0.0',
         'twine>=3.1.1,<4.0.0',
         'wheel>=0.33.6,<0.34.0']}

entry_points = \
{'console_scripts': ['sdss = sdsstools.cli:main']}

setup_kwargs = {
    'name': 'sdsstools',
    'version': '0.1.11',
    'description': 'Small tools for SDSS products',
    'long_description': '# sdsstools\n\n![Versions](https://img.shields.io/badge/python->3.7-blue)\n[![PyPI version](https://badge.fury.io/py/sdsstools.svg)](https://badge.fury.io/py/sdsstools)\n\n`sdsstools` provides several common tools for logging, configuration handling, version parsing, packaging, etc. It\'s main purpose is to consolidate some of the utilities originally found in the [python_template](https://github.com/sdss/python_template), allowing them to become dependencies that can be updated.\n\n**This is not intended to be a catch-all repository for astronomical tools.** `sdsstools` itself aims to have minimal dependencies (i.e., mainly the Python standard library and setuptools).\n\n## Using sdsstools\n\nTo use sdsstools simply install it with\n\n```console\npip install sdsstools\n```\n\nMost likely, you\'ll want to include sdsstools as a dependency for your library. To do so, either add to your `setup.cfg`\n\n```ini\n[options]\ninstall_requires =\n    sdsstools>=0.1.0\n```\n\n(this is equivalent of passing `install_requires=[\'sdsstools>=0.1.0\']` to `setuptools.setup`), or if you are using [poetry](https://poetry.eustace.io/) run `poetry add sdsstools`, which should add this line to your `pyproject.toml`\n\n```toml\n[tool.poetry.dependencies]\nsdsstools = { version="^0.1.0" }\n```\n\n## Logging\n\nsdsstools includes the [sdsstools.logger.SDSSLogger](https://github.com/sdss/sdsstools/blob/f30e00f527660fe8627e33a7f931b44410b0ff06/src/sdsstools/logger.py#L107) class, which provides a wrapper around the standard Python [logging](https://docs.python.org/3/library/logging.html) module. `SDSSLoger` provides the following features:\n\n- A console handler (accessible via the `.sh` attribute) with nice colouring.\n- Automatic capture of warnings and exceptions, which are formatted and redirected to the logger. For the console handler, this means that once the logger has been created, all warnings and exceptions are output normally but are clearer and more aesthetic.\n- A [TimedRotatingFileHandler](https://docs.python.org/3.8/library/logging.handlers.html#logging.handlers.TimedRotatingFileHandler) (accessible via the `.fh` attribute) that rotates at midnight UT, with good formatting.\n\nTo get a new logger for your application, simply do\n\n```python\nfrom sdsstools.logger import get_logger\n\nNAME = \'myrepo\'\nlog = get_logger(NAME)\n```\n\nThe file logger is disabled by default and can be started by calling `log.start_file_logger(path)`. By default a `TimedRotatingFileHandler` is created. If you want a normal `FileHandler` use `rotate=False`. The file mode defaults to `mode=\'a\'` (append).\n\n## Configuration\n\nThe `sdsstools.configuration` module contains several utilities to deal with configuration files. The most useful one is [get_config](https://github.com/sdss/sdsstools/blob/d3d337953a37aaff9c38fead76b08b414164775a/src/sdsstools/configuration.py#L40), which allows to read a YAML configuration file. For example\n\n```python\nfrom sdsstools.configuration import get_config\n\nNAME = \'myrepo\'\nconfig = get_config(NAME, allow_user=True)\n```\n\n`get_config` assumes that the file is located in `etc/<NAME>.yml` relative from the file that calls `get_config`, but that can be changed by passing `config_file=<config-file-path>`. Additionally, if `allow_user=True` and a file exists in `~/.config/sdss/<NAME>.yaml`, this file is read and merged with the default configuration, overriding any parameter that is present in the user file. This allows to create a default configuration that lives with the library but that can be overridden by a user.\n\nIn addition to the (recommended) location `~/.config/sdss/<NAME>.yaml`, `get_config` also looks for user configuration files in `~/.config/sdss/<NAME>.yml`, `~/.config/sdss/<NAME>/<NAME>.y(a)ml`, and `~/.<NAME>/<NAME>.y(a)ml`.\n\nAdditionally, `sdsstools.configuration` includes two other tools, `merge_config`, that allows to merge dictionaries recursively, and `read_yaml_file`, to read a YAML file.\n\n## Metadata\n\nsdsscore provides tools to locate and parse metadata files (`pyproject.toml`, `setup.cfg`, `setup.py`). `get_metadata_files` locates the path of the metadata file relative to a given `path`. `get_package_version` tries to find the version of the package by looking for a version string in the metadata file or in the egg/wheel metadata file, if the package has been installed. To use it\n\n```python\nfrom sdsstools.metadata import get_package_version\n\n__version__ = get_package_version(path=__file__, package_name=\'sdss-camera\') or \'dev\'\n```\n\nThis will try to find and parse the version from the metadata file (we pass `__file__` to indicate where to start looking); if that fails, it will try to get the version from the installed package `sdss-camera`. If all fails, it will set the fallback version `\'dev\'`.\n\n## Command Line Interface\n\n`sdsstools` provides the command line tool `sdss`, which is just a thin wrapper around some commonly used [Invoke](https://www.pyinvoke.org/) tasks. Because `sdsstools` tries very hard not to install unnecessary dependencies, and the CLI should only be useful for development, you\'ll need to run `pip install sdsscore[dev]` to be able to use `sdss`.\n\n`sdss` provides the following tasks\n\n| Task | Options | Description |\n| --- | --- | --- |\n| clean | | Removes files produces during build and packaging |\n| deploy | --test | Builds and deploys to PyPI (or the test server) |\n| install-deps | --extras | Installs dependencies from a `setup.cfg` file |\n| docs.build | --target | Builds the Sphinx documentation |\n| docs.show | --target | Shows the documentation in the browser |\n| docs.clean | --target | Cleans the documentation build |\n\n`sdss` assumes that the documentation lives in `docs/sphinx` relative to the root of the repository. This can be changed by setting the `sphinx.target` configuration in an `invoke.yaml` file, for example\n\n```yaml\nsphinx:\n    target: docs\n```\n\n## Bundled packages\n\nFor convenience, `sdsstools` bundles the following products:\n\n- A copy of [releases](https://github.com/bitprophet/releases) that fixes some issues with recent versions of `semantic-version`.\n- A copy of [toml](https://github.com/uiri/toml) to read TOML files (used by the metadata submodule).\n\nYou can access them directly from the top-level namespace, `sdsstools.toml`, `sdsstools.releases`. To use `releases` with sphinx, simply add the following to your `config.py`\n\n```python\nextensions += [\'sdsstools.releases\']\n```\n',
    'author': 'José Sánchez-Gallego',
    'author_email': 'gallegoj@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sdss/sdsstools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import itertools
import re
import sys
from functools import partial

from docutils import nodes, utils
from docutils.parsers.rst import roles
import six

from .models import Issue, ISSUE_TYPES, Release, Version, Spec
from .line_manager import LineManager
from ._version import __version__


def _log(txt, config):
    """
    Log debug output if debug setting is on.

    Intended to be partial'd w/ config at top of functions. Meh.
    """
    if config.releases_debug:
        sys.stderr.write(str(txt) + "\n")
        sys.stderr.flush()


def issue_nodelist(name, identifier=None):
    which = '[<span style="color: #%s;">%s</span>]' % (
        ISSUE_TYPES[name], name.capitalize()
    )
    signifier = [nodes.raw(text=which, format='html')]
    id_nodelist = [nodes.inline(text=" "), identifier] if identifier else []
    trail = [] if identifier else [nodes.inline(text=" ")]
    return signifier + id_nodelist + [nodes.inline(text=":")] + trail


release_line_re = re.compile(r'^(\d+\.\d+)\+$') # e.g. '1.2+'

def scan_for_spec(keyword):
    """
    Attempt to return some sort of Spec from given keyword value.

    Returns None if one could not be derived.
    """
    # Both 'spec' formats are wrapped in parens, discard
    keyword = keyword.lstrip('(').rstrip(')')
    # First, test for intermediate '1.2+' style
    matches = release_line_re.findall(keyword)
    if matches:
        return Spec(">={}".format(matches[0]))
    # Failing that, see if Spec can make sense of it
    try:
        return Spec(keyword)
    # I've only ever seen Spec fail with ValueError.
    except ValueError:
        return None


def issues_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Use: :issue|bug|feature|support:`ticket_number`

    When invoked as :issue:, turns into just a "#NN" hyperlink to
    `releases_issue_uri`.

    When invoked otherwise, turns into "[Type] <#NN hyperlink>: ".

    Spaces present in the "ticket number" are used as fields for keywords
    (major, backported) and/or specs (e.g. '>=1.0'). This data is removed &
    used when constructing the object.

    May give a 'ticket number' of ``-`` or ``0`` to generate no hyperlink.
    """
    parts = utils.unescape(text).split()
    issue_no = parts.pop(0)
    # Lol @ access back to Sphinx
    config = inliner.document.settings.env.app.config
    if issue_no not in ('-', '0'):
        ref = None
        if config.releases_issue_uri:
            # TODO: deal with % vs .format()
            ref = config.releases_issue_uri % issue_no
        elif config.releases_github_path:
            ref = "https://github.com/{}/issues/{}".format(
                config.releases_github_path, issue_no)
        # Only generate a reference/link if we were able to make a URI
        if ref:
            identifier = nodes.reference(
                rawtext, '#' + issue_no, refuri=ref, **options
            )
        # Otherwise, just make it regular text
        else:
            identifier = nodes.raw(
                rawtext=rawtext, text='#' + issue_no, format='html',
                **options
            )
    else:
        identifier = None
        issue_no = None # So it doesn't gum up dupe detection later
    # Additional 'new-style changelog' stuff
    if name in ISSUE_TYPES:
        nodelist = issue_nodelist(name, identifier)
        spec = None
        keyword = None
        # TODO: sanity checks re: e.g. >2 parts, >1 instance of keywords, >1
        # instance of specs, etc.
        for part in parts:
            maybe_spec = scan_for_spec(part)
            if maybe_spec:
                spec = maybe_spec
            else:
                if part in ('backported', 'major'):
                    keyword = part
                else:
                    err = "Gave unknown keyword {!r} for issue no. {}"
                    raise ValueError(err.format(keyword, issue_no))
        # Create temporary node w/ data & final nodes to publish
        node = Issue(
            number=issue_no,
            type_=name,
            nodelist=nodelist,
            backported=(keyword == 'backported'),
            major=(keyword == 'major'),
            spec=spec,
        )
        return [node], []
    # Return old style info for 'issue' for older changelog entries
    else:
        return [identifier], []


def release_nodes(text, slug, date, config):
    # Doesn't seem possible to do this "cleanly" (i.e. just say "make me a
    # title and give it these HTML attributes during render time) so...fuckit.
    # We were already doing fully raw elements elsewhere anyway. And who cares
    # about a PDF of a changelog? :x
    uri = None
    if config.releases_release_uri:
        # TODO: % vs .format()
        uri = config.releases_release_uri % slug
    elif config.releases_github_path:
        uri = "https://github.com/{}/tree/{}".format(
            config.releases_github_path, slug)
    # Only construct link tag if user actually configured release URIs somehow
    if uri:
        link = '<a class="reference external" href="{}">{}</a>'.format(
            uri, text,
        )
    else:
        link = text
    datespan = ''
    if date:
        datespan = ' <span style="font-size: 75%;">{}</span>'.format(date)
    header = '<h2 style="margin-bottom: 0.3em;">{}{}</h2>'.format(
        link, datespan)
    return nodes.section('',
        nodes.raw(rawtext='', text=header, format='html'),
        ids=[text]
    )


year_arg_re = re.compile(r'^(.+?)\s*(?<!\x00)<(.*?)>$', re.DOTALL)


def release_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Invoked as :release:`N.N.N <YYYY-MM-DD>`.

    Turns into useful release header + link to GH tree for the tag.
    """
    # Make sure year has been specified
    match = year_arg_re.match(text)
    if not match:
        msg = inliner.reporter.error("Must specify release date!")
        return [inliner.problematic(rawtext, rawtext, msg)], [msg]
    number, date = match.group(1), match.group(2)
    # Lol @ access back to Sphinx
    config = inliner.document.settings.env.app.config
    nodelist = [release_nodes(number, number, date, config)]
    # Return intermediate node
    node = Release(number=number, date=date, nodelist=nodelist)
    return [node], []


def generate_unreleased_entry(header, line, issues, manager, app):
    log = partial(_log, config=app.config)
    nodelist = [release_nodes(
        header,
        # TODO: should link to master for newest family and...what
        # exactly, for the others? Expectation isn't necessarily to
        # have a branch per family? Or is there? Maybe there must be..
        'master',
        None,
        app.config
    )]
    log("Creating {!r} faux-release with {!r}".format(line, issues))
    return {
        'obj': Release(number=line, date=None, nodelist=nodelist),
        'entries': issues,
    }


def append_unreleased_entries(app, manager, releases):
    """
    Generate new abstract 'releases' for unreleased issues.

    There's one for each combination of bug-vs-feature & major release line.

    When only one major release line exists, that dimension is ignored.
    """
    for family, lines in six.iteritems(manager):
        for type_ in ('bugfix', 'feature'):
            bucket = 'unreleased_{}'.format(type_)
            if bucket not in lines: # Implies unstable prehistory + 0.x fam
                continue
            issues = lines[bucket]
            fam_prefix = "{}.x ".format(family) if len(manager) > 1 else ""
            header = "Next {}{} release".format(fam_prefix, type_)
            line = "unreleased_{}.x_{}".format(family, type_)
            releases.append(
                generate_unreleased_entry(header, line, issues, manager, app)
            )


def reorder_release_entries(releases):
    """
    Mutate ``releases`` so the entrylist in each is ordered by feature/bug/etc.
    """
    order = {'feature': 0, 'bug': 1, 'support': 2}
    for release in releases:
        entries = release['entries'][:]
        release['entries'] = sorted(entries, key=lambda x: order[x.type])


def construct_entry_with_release(focus, issues, manager, log, releases, rest):
    """
    Releases 'eat' the entries in their line's list and get added to the
    final data structure. They also inform new release-line 'buffers'.
    Release lines, once the release obj is removed, should be empty or a
    comma-separated list of issue numbers.
    """
    log("release for line %r" % focus.minor)
    # Check for explicitly listed issues first
    explicit = None
    if rest[0].children:
        explicit = [x.strip() for x in rest[0][0].split(',')]
    # Do those by themselves since they override all other logic
    if explicit:
        log("Explicit issues requested: %r" % (explicit,))
        # First scan global issue dict, dying if not found
        missing = [i for i in explicit if i not in issues]
        if missing:
            raise ValueError(
                "Couldn't find issue(s) #{} in the changelog!".format(
                    ', '.join(missing)))
        # Obtain the explicitly named issues from global list
        entries = []
        for i in explicit:
            for flattened_issue_item in itertools.chain(issues[i]):
                entries.append(flattened_issue_item)
        # Create release
        log("entries in this release: %r" % (entries,))
        releases.append({
            'obj': focus,
            'entries': entries,
        })
        # Introspect these entries to determine which buckets they should get
        # removed from (it's not "all of them"!)
        for obj in entries:
            if obj.type == 'bug':
                # Major bugfix: remove from unreleased_feature
                if obj.major:
                    log("Removing #%s from unreleased" % obj.number)
                    # TODO: consider making a LineManager method somehow
                    manager[focus.family]['unreleased_feature'].remove(obj)
                # Regular bugfix: remove from bucket for this release's
                # line + unreleased_bugfix
                else:
                    if obj in manager[focus.family]['unreleased_bugfix']:
                        log("Removing #%s from unreleased" % obj.number)
                        manager[focus.family]['unreleased_bugfix'].remove(obj)
                    if obj in manager[focus.family][focus.minor]:
                        log("Removing #%s from %s" % (obj.number, focus.minor))
                        manager[focus.family][focus.minor].remove(obj)
            # Regular feature/support: remove from unreleased_feature
            # Backported feature/support: remove from bucket for this
            # release's line (if applicable) + unreleased_feature
            else:
                log("Removing #%s from unreleased" % obj.number)
                manager[focus.family]['unreleased_feature'].remove(obj)
                if obj in manager[focus.family].get(focus.minor, []):
                    manager[focus.family][focus.minor].remove(obj)

    # Implicit behavior otherwise
    else:
        # Unstable prehistory -> just dump 'unreleased' and continue
        if manager.unstable_prehistory:
            # TODO: need to continue making LineManager actually OO, i.e. do
            # away with the subdicts + keys, move to sub-objects with methods
            # answering questions like "what should I give you for a release"
            # or whatever
            log("in unstable prehistory, dumping 'unreleased'")
            releases.append({
                'obj': focus,
                # NOTE: explicitly dumping 0, not focus.family, since this
                # might be the last pre-historical release and thus not 0.x
                'entries': manager[0]['unreleased'][:],
            })
            manager[0]['unreleased'] = []
            # If this isn't a 0.x release, it signals end of prehistory, make a
            # new release bucket (as is also done below in regular behavior).
            # Also acts like a sentinel that prehistory is over.
            if focus.family != 0:
                manager[focus.family][focus.minor] = []
        # Regular behavior from here
        else:
            # New release line/branch detected. Create it & dump unreleased
            # features.
            if focus.minor not in manager[focus.family]:
                log("not seen prior, making feature release & bugfix bucket")
                manager[focus.family][focus.minor] = []
                # TODO: this used to explicitly say "go over everything in
                # unreleased_feature and dump if it's feature, support or major
                # bug". But what the hell else would BE in unreleased_feature?
                # Why not just dump the whole thing??
                #
                # Dump only the items in the bucket whose family this release
                # object belongs to, i.e. 1.5.0 should only nab the 1.0
                # family's unreleased feature items.
                releases.append({
                    'obj': focus,
                    'entries': manager[focus.family]['unreleased_feature'][:],
                })
                manager[focus.family]['unreleased_feature'] = []

            # Existing line -> empty out its bucket into new release.
            # Skip 'major' bugs as those "belong" to the next release (and will
            # also be in 'unreleased_feature' - so safe to nuke the entire
            # line)
            else:
                log("pre-existing, making bugfix release")
                # TODO: as in other branch, I don't get why this wasn't just
                # dumping the whole thing - why would major bugs be in the
                # regular bugfix buckets?
                entries = manager[focus.family][focus.minor][:]
                releases.append({'obj': focus, 'entries': entries})
                manager[focus.family][focus.minor] = []
                # Clean out the items we just released from
                # 'unreleased_bugfix'.  (Can't nuke it because there might
                # be some unreleased bugs for other release lines.)
                for x in entries:
                    if x in manager[focus.family]['unreleased_bugfix']:
                        manager[focus.family]['unreleased_bugfix'].remove(x)


def construct_entry_without_release(focus, issues, manager, log, rest):
    # Handle rare-but-valid non-issue-attached line items, which are
    # always bugs. (They are their own description.)
    if not isinstance(focus, Issue):
        # First, sanity check for potential mistakes resulting in an issue node
        # being buried within something else.
        buried = focus.traverse(Issue)
        if buried:
            msg = """
Found issue node ({!r}) buried inside another node:

{}

Please double-check your ReST syntax! There is probably text in the above
output that will show you which part of your changelog to look at.

For example, indentation problems can accidentally generate nested definition
lists.
"""
            raise ValueError(msg.format(buried[0], str(buried[0].parent)))
        # OK, it looks legit - make it a bug.
        log("Found line item w/ no real issue object, creating bug")
        nodelist = issue_nodelist('bug')
        # Skip nodelist entirely if we're in unstable prehistory -
        # classification doesn't matter there.
        if manager.unstable_prehistory:
            nodelist = []
        # Undo the 'pop' from outer scope. TODO: rework things so we don't have
        # to do this dumb shit uggggh
        rest[0].insert(0, focus)
        focus = Issue(
            type_='bug',
            nodelist=nodelist,
            description=rest,
        )
    else:
        focus.attributes['description'] = rest

    # Add to global list (for use by explicit releases) or die trying
    issues[focus.number] = issues.get(focus.number, []) + [focus]

    # Add to per-release bugfix lines and/or unreleased bug/feature buckets, as
    # necessary.
    # TODO: suspect all of add_to_manager can now live in the manager; most of
    # Release's methods should probably go that way
    if manager.unstable_prehistory:
        log("Unstable prehistory -> adding to 0.x unreleased bucket")
        manager[0]['unreleased'].append(focus)
    else:
        log("Adding to release line manager")
        focus.add_to_manager(manager)


def handle_upcoming_major_release(entries, manager):
    # Short-circuit if the future holds nothing for us
    if not entries:
        return
    # Short-circuit if we're in the middle of a block of releases, only the
    # last release before a bunch of issues, should be taking any action.
    if isinstance(entries[0], Release):
        return
    # Iterate through entries til we find the next Release or set of Releases
    next_releases = []
    for index, obj in enumerate(entries):
        if isinstance(obj, Release):
            next_releases.append(obj)
        # Non-empty next_releases + encountered a non-release = done w/ release
        # block.
        elif next_releases:
            break
    # Examine result: is a major release present? If so, add its major number
    # to the line manager!
    for obj in next_releases:
        # TODO: update when Release gets tied closer w/ Version
        version = Version(obj.number)
        if version.minor == 0 and version.patch == 0:
            manager.add_family(obj.family)


def handle_first_release_line(entries, manager):
    """
    Set up initial line-manager entry for first encountered release line.

    To be called at start of overall process; afterwards, subsequent major
    lines are generated by `handle_upcoming_major_release`.
    """
    # It's remotely possible the changelog is totally empty...
    if not entries:
        return
    # Obtain (short-circuiting) first Release obj.
    first_release = None
    for obj in entries:
        if isinstance(obj, Release):
            first_release = obj
            break
    # It's also possible it's non-empty but has no releases yet.
    if first_release:
        manager.add_family(obj.family)
    # If God did not exist, man would be forced to invent him.
    else:
        manager.add_family(0)


def construct_releases(entries, app):
    log = partial(_log, config=app.config)
    # Walk from back to front, consuming entries & copying them into
    # per-release buckets as releases are encountered. Store releases in order.
    releases = []
    # Release lines, to be organized by major releases, then by major+minor,
    # alongside per-major-release 'unreleased' bugfix/feature buckets.
    # NOTE: With exception of unstable_prehistory=True, which triggers use of a
    # separate, undifferentiated 'unreleased' bucket (albeit still within the
    # '0' major line family).
    manager = LineManager(app)
    # Also keep a master hash of issues by number to detect duplicates & assist
    # in explicitly defined release lists.
    issues = {}

    reversed_entries = list(reversed(entries))
    # For the lookahead, so we're not doing this stripping O(n) times.
    # TODO: probs just merge the two into e.g. a list of 2-tuples of "actual
    # entry obj + rest"?
    stripped_entries = [x[0][0] for x in reversed_entries]
    # Perform an initial lookahead to prime manager with the 1st major release
    handle_first_release_line(stripped_entries, manager)
    # Start crawling...
    for index, obj in enumerate(reversed_entries):
        # Issue object is always found in obj (LI) index 0 (first, often only
        # P) and is the 1st item within that (index 0 again).
        # Preserve all other contents of 'obj'.
        focus = obj[0].pop(0)
        rest = obj
        log(repr(focus))
        # Releases 'eat' the entries in their line's list and get added to the
        # final data structure. They also inform new release-line 'buffers'.
        # Release lines, once the release obj is removed, should be empty or a
        # comma-separated list of issue numbers.
        if isinstance(focus, Release):
            construct_entry_with_release(
                focus, issues, manager, log, releases, rest
            )
            # After each release is handled, look ahead to see if we're
            # entering "last stretch before a major release". If so,
            # pre-emptively update the line-manager so upcoming features are
            # correctly sorted into that major release by default (re: logic in
            # Release.add_to_manager)
            handle_upcoming_major_release(
                stripped_entries[index + 1:], manager
            )

        # Entries get copied into release line buckets as follows:
        # * Features and support go into 'unreleased_feature' for use in new
        # feature releases.
        # * Bugfixes go into all release lines (so they can be printed in >1
        # bugfix release as appropriate) as well as 'unreleased_bugfix' (so
        # they can be displayed prior to release'). Caveats include bugs marked
        # 'major' (they go into unreleased_feature instead) or with 'N.N+'
        # (meaning they only go into release line buckets for that release and
        # up.)
        # * Support/feature entries marked as 'backported' go into all
        # release lines as well, on the assumption that they were released to
        # all active branches.
        # * The 'rest' variable (which here is the bug description, vitally
        # important!) is preserved by stuffing it into the focus (issue)
        # object - it will get unpacked by construct_nodes() later.
        else:
            construct_entry_without_release(focus, issues, manager, log, rest)

    if manager.unstable_prehistory:
        releases.append(generate_unreleased_entry(
            header="Next release",
            line="unreleased",
            issues=manager[0]['unreleased'],
            manager=manager,
            app=app,
        ))
    else:
        append_unreleased_entries(app, manager, releases)

    reorder_release_entries(releases)

    return releases, manager


def construct_nodes(releases):
    result = []
    # Reverse the list again so the final display is newest on top
    for d in reversed(releases):
        if not d['entries']:
            continue
        obj = d['obj']
        entries = []
        for entry in d['entries']:
            # Use nodes.Node.deepcopy to deepcopy the description
            # node.  If this is not done, multiple references to the same
            # object (e.g. a reference object in the description of #649, which
            # is then copied into 2 different release lists) will end up in the
            # doctree, which makes subsequent parse steps very angry (index()
            # errors).
            desc = entry['description'].deepcopy()
            # Additionally, expand any other issue roles found in the
            # description - sometimes we refer to related issues inline. (They
            # can't be left as issue() objects at render time since that's
            # undefined.)
            # Use [:] slicing to avoid mutation during the loops.
            for index, node in enumerate(desc[:]):
                for subindex, subnode in enumerate(node[:]):
                    if isinstance(subnode, Issue):
                        lst = subnode['nodelist']
                        desc[index][subindex:subindex + 1] = lst
            # Rework this entry to insert the now-rendered issue nodes in front
            # of the 1st paragraph of the 'description' nodes (which should be
            # the preserved LI + nested paragraph-or-more from original
            # markup.)
            # FIXME: why is there no "prepend a list" method?
            for node in reversed(entry['nodelist']):
                desc[0].insert(0, node)
            entries.append(desc)
        # Entry list
        list_ = nodes.bullet_list('', *entries)
        # Insert list into release nodelist (as it's a section)
        obj['nodelist'][0].append(list_)
        # Release header
        header = nodes.paragraph('', '', *obj['nodelist'])
        result.extend(header)
    return result


class BulletListVisitor(nodes.NodeVisitor):
    def __init__(self, document, app):
        nodes.NodeVisitor.__init__(self, document)
        self.found_changelog = False
        self.app = app

    def visit_bullet_list(self, node):
        # The first found bullet list (which should be the first one at the top
        # level of the document) is the changelog.
        if not self.found_changelog:
            self.found_changelog = True
            # Walk + parse into release mapping
            releases, _ = construct_releases(node.children, self.app)
            # Construct new set of nodes to replace the old, and we're done
            node.replace_self(construct_nodes(releases))

    def unknown_visit(self, node):
        pass


def generate_changelog(app, doctree):
    # Don't scan/mutate documents that don't match the configured document name
    # (which by default is ['changelog.rst', ]).
    if app.env.docname not in app.config.releases_document_name:
        return

    # Find the first bullet-list node & replace it with our organized/parsed
    # elements.
    changelog_visitor = BulletListVisitor(doctree, app)
    doctree.walk(changelog_visitor)


def setup(app):
    for key, default in (
        # Issue base URI setting: releases_issue_uri
        # E.g. 'https://github.com/fabric/fabric/issues/'
        ('issue_uri', None),
        # Release-tag base URI setting: releases_release_uri
        # E.g. 'https://github.com/fabric/fabric/tree/'
        ('release_uri', None),
        # Convenience Github version of above
        ('github_path', None),
        # Which document to use as the changelog
        ('document_name', ['changelog']),
        # Debug output
        ('debug', False),
        # Whether to enable linear history during 0.x release timeline
        # TODO: flip this to True by default in our 2.0 release
        ('unstable_prehistory', False),
    ):
        app.add_config_value(
            name='releases_{}'.format(key), default=default, rebuild='html'
        )
    # if a string is given for `document_name`, convert it to a list
    # done to maintain backwards compatibility
    # https://stackoverflow.com/questions/1303243/how-to-find-out-if-a-python-object-is-a-string
    PY2 = sys.version_info[0] == 2
    if PY2:
        string_types = (basestring,)
    else:
        string_types = (str,)

    if isinstance(app.config.releases_document_name, string_types):
        app.config.releases_document_name = [app.config.releases_document_name]

    # Register intermediate roles
    for x in list(ISSUE_TYPES) + ['issue']:
        add_role(app, x, issues_role)
    add_role(app, 'release', release_role)
    # Hook in our changelog transmutation at appropriate step
    app.connect('doctree-read', generate_changelog)

    # identifies the version of our extension
    return {'version': __version__}


def add_role(app, name, role_obj):
    # This (introspecting docutils.parser.rst.roles._roles) is the same trick
    # Sphinx uses to emit warnings about double-registering; it's a PITA to try
    # and configure the app early on so it doesn't emit those warnings, so we
    # instead just...don't double-register. Meh.
    if name not in roles._roles:
        app.add_role(name, role_obj)
